#include "hooks.h"
#include "hipp.h"
#include "securitycheck.h"
#include "pstate.h"
#include "printlog.h"
#include "utilities.h"
#include "filesystem.h"
#include "apply_exception.h"

struct security_operations hipp_ops;

int hipp_task_setuid(uid_t id0, uid_t id1, uid_t id2, int flags)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	
	if (pstate == NULL) {
		print_log(ERR_INF, "null error in [hipp_task_setuid]\n");
		return 0;
	}

	if (setuid_check(id0, id1, id2, pstate, 7) != 0) {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to setuid.");
		print_log(CTN_INF, "\n");
	}
	return 0;
}

int hipp_task_setgid(gid_t id0, gid_t id1, gid_t id2, int flags)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	
	if (pstate == NULL) {
		print_log(ERR_INF, "null error in [hipp_task_setgid]\n");
		return 0;
	}
	
	if (setuid_check(id0, id1, id2, pstate, 6) != 0) {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to setgid.");
		print_log(CTN_INF, "\n");
	}
	return 0;
}

int hipp_task_post_setuid(uid_t old_ruid, uid_t old_euid, uid_t old_suid, int flags)
{
	int cur_uid;
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	int rc = cap_task_post_setuid(old_ruid, old_euid, old_suid, flags);
	
	if (pstate == NULL) {
		return rc;
	}
	
	cur_uid = get_huid(current);
	pstate->huid = update_huid(pstate->huid, cur_uid); 
	return rc;
}

void hipp_bprm_apply_creds(struct linux_binprm *bprm, int unsafe)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	struct dentry* dentry;
	char* path;

	cap_bprm_apply_creds(bprm, unsafe);
	
	if (bprm == NULL || bprm->filename == NULL || pstate == NULL) {
		print_log(ERR_INF, "null error in [bprm_apply_creds]\n");
		return;
	}
	
	if (pstate->path != NULL && strcmp(pstate->path, UPGRADE_INTEGRITY_UTILITY_NAME) == 0) {
		return;
	}

	path = kmalloc(MAX_PATH_LEN, GFP_KERNEL);
	if (path == NULL) {
		print_log(ERR_INF, "null error in [hipp_bprm_apply_creds].\n");
		return;
	}
	
	dentry = get_dentry_from_path(bprm->filename);
	if (dentry == NULL) {
		strcpy(path, bprm->filename);
	}
	else {
		if (get_path_from_dentry(dentry, path) == -1) {
			strcpy(path, bprm->filename);
		}
	}
	
	print_prs(PGS_INF, current);
	print_log(CTN_INF, "exec");
	print_log(CTN_INF, path);
	print_log(CTN_INF, "\n");

	exec_program(pstate, path, dentry);
	//print_apply_exception(&pstate->apply_exceptions);
	
	//for debug
	if (strcmp(bprm->filename, "/bin/bash") == 0) {
	//	pstate->integrity_state.integrity_level = LOW_INTEGRITY;
	}
	kfree(path);
}

int hipp_inode_create(struct inode *dir, struct dentry *dentry, int mode)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (dir == NULL || dentry == NULL || pstate == NULL) {
		return 0;
	}
	
	if (create_check(dentry, pstate) == 0) {
		
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to create the inode");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_link(struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (dir == NULL || old_dentry == NULL || new_dentry == NULL || pstate == NULL) {
		return 0;
	}
	
	if (create_link_check(old_dentry, new_dentry, pstate) == 0) {
		
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to create hard link ");
		print_den(CTN_INF, new_dentry);
		print_log(CTN_INF, " to ");
		print_den(CTN_INF, old_dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_unlink(struct inode *dir, struct dentry *dentry)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (dentry == NULL || dir == NULL || pstate == NULL) {
		return 0;
	}
	
	if (unlink_check(dentry, pstate) == 0) {
		
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to remove hard link ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_symlink(struct inode *dir, struct dentry *dentry, const char *old_name)
{
	struct dentry* old_dentry;
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (old_name == NULL || dentry == NULL || dir == NULL || pstate == NULL) {
		return 0;
	}
	
	old_dentry = get_dentry_from_path(old_name);
	
	if (old_dentry == NULL) {
		return 0;
	}
	
	if (create_link_check(old_dentry, dentry, pstate) == 0) {
		
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to create soft link ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, " to ");
		print_log(CTN_INF, old_name);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (dentry == NULL || dir == NULL || pstate == NULL) {
		return 0;
	}
	
	if (create_check(dentry, pstate) == 0) {
		
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to create the directory ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (dentry == NULL || dir == NULL || pstate == NULL) {
		return 0;
	}
	
	if (unlink_check(dentry, pstate) == 0) {
		
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to remove the directory ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_rename(struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (old_dentry == NULL || new_dentry == NULL || old_dir == NULL || new_dir == NULL || pstate == NULL) {
		return 0;
	}
	
	if (rename_check(old_dentry, new_dentry, pstate) == 0) {
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to rename from ");
		print_den(CTN_INF, old_dentry);
		print_log(CTN_INF, " to ");
		print_den(CTN_INF, new_dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_permission(struct inode *inode, int mask, struct nameidata *nd)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (mask == 0 || inode == NULL || nd == NULL || nd->dentry == NULL || pstate == NULL) { 
		return 0;
	}

	if (mask & MAY_READ) {
		if (read_check(nd->dentry, pstate) == 0) {
			read_from(nd->dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to read from inode ");
			print_den(CTN_INF, nd->dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}
	
	if (mask & MAY_WRITE || mask & MAY_APPEND) {
		if (write_check(nd->dentry, pstate) == 0) {
			write_to(nd->dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to write to inode ");
			print_den(CTN_INF, nd->dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}

	if (mask & MAY_EXEC) {
		if (exec_check(nd->dentry, pstate) == 0) {
			exec_from(nd->dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to exec inode ");
			print_den(CTN_INF, nd->dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}

	return 0;
}

int hipp_inode_setattr(struct dentry *dentry, struct iattr *iattr)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	int sticky = 0;

	if (dentry == NULL || iattr == NULL || dentry->d_inode == NULL || pstate == NULL) {
		return 0;
	}
	
	if (set_permissions_check(dentry, pstate, iattr->ia_mode, iattr->ia_uid, &sticky) == 0) {
		if (sticky == 1) {
			iattr->ia_mode = (iattr->ia_mode | S_ISVTX);
		}
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to set attribute ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_setxattr(struct dentry *dentry, char *name, void *value, size_t size, int flags)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	int rc = cap_inode_setxattr(dentry, name, value, size, flags);	
	if (rc < 0) {
		return rc;
	}
	
	if (dentry == NULL || pstate == NULL) {
		return 0;
	}
	
	if (write_check(dentry, pstate) == 0) {
		write_to(dentry, pstate);
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to set extra attribute ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_inode_removexattr(struct dentry *dentry, char *name)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	int rc = cap_inode_removexattr(dentry, name);
	if (rc < 0) {
		return rc;
	}
	
	if (dentry == NULL || pstate == NULL) {
		return 0;
	}
	
	if (write_check(dentry, pstate) == 0) {
		write_to(dentry, pstate);
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to remove extra attribute ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

int hipp_file_permission(struct file *file, int mask)
{
	struct inode *inode;
	struct dentry *dentry;

	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (file == NULL || file->f_dentry == NULL || file->f_dentry->d_inode == NULL || pstate == NULL) {
		return 0;
	}
	
	dentry = file->f_dentry;
	inode = file->f_dentry->d_inode;

	if (mask & MAY_READ) {
		if (read_check(dentry, pstate) == 0) {
			read_from(dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to read from file ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}
	
	if (mask & MAY_WRITE || mask & MAY_APPEND) {
		if (write_check(dentry, pstate) == 0) {
			write_to(dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to write to file ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}

	if (mask & MAY_EXEC) {
		if (exec_check(dentry, pstate) == 0) {
			exec_from(dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to exec file ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}

	return 0;
}

int hipp_file_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
/*
	struct inode *inode;
	struct dentry *dentry;

	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (file == NULL || file->f_dentry == NULL || file->f_dentry->d_inode == NULL || pstate == NULL) {
		return 0;
	}
	
	dentry = file->f_dentry;
	inode = file->f_dentry->d_inode;
	
	if (write_check(dentry, pstate) == 0) {
		write_to(dentry, pstate);
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to io control file ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
*/	return 0;
}

int hipp_file_fcntl(struct file *file, unsigned int cmd, unsigned long arg)
{
	// TODO:
	return 0;
}

int file_map_prot_check(struct file *file, unsigned long prot, int shared)
{
	struct inode *inode;
	struct dentry *dentry;

	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	if (file == NULL || file->f_dentry == NULL || file->f_dentry->d_inode == NULL || pstate == NULL) {
		return 0;
	}
	
	dentry = file->f_dentry;
	inode = file->f_dentry->d_inode;
	
	if (read_check(dentry, pstate) == 0) {
		read_from(dentry, pstate);
	}
	else {
		print_prs(EXP_INF, current);
		print_log(CTN_INF, " failed to mmap the file (r) ");
		print_den(CTN_INF, dentry);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
		
	if (shared && (prot & PROT_WRITE)) {
		if (write_check(dentry, pstate) == 0) {
			write_to(dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to mmap the file (w) ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}
	
	if (prot & PROT_EXEC) {
		if (exec_check(dentry, pstate) == 0) {
			exec_from(dentry, pstate);
		}
		else {
			print_prs(EXP_INF, current);
			print_log(CTN_INF, " failed to mmap the file (e) ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");
			
			return error_code;
		}
	}
	return 0;
}

int hipp_file_mmap(struct file *file, unsigned long reqprot, unsigned long prot, unsigned long flags)
{
	return file_map_prot_check(file, prot, (flags & MAP_TYPE) == MAP_SHARED);
}

int hipp_file_mprotect(struct vm_area_struct *vma, unsigned long reqprot, unsigned long prot)
{
	if (vma == NULL) {
		return 0;
	}
	
	return file_map_prot_check(vma->vm_file, prot, vma->vm_flags & VM_SHARED);
}

int hipp_capable(struct task_struct *tsk, int cap)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	char buf[128];

	int rc = cap_capable(tsk, cap);	
	if (rc < 0) {
		return rc;
	}
	
	if (capability_check(cap, pstate) == 0) {
		
	}
	else {
		sprintf(buf, " failed to enforce capability %d.\n", cap);
		print_prs(EXP_INF, current);
		print_log(CTN_INF, buf);		
	}
	
	return 0;
}

int hipp_task_alloc_security(struct task_struct *tsk)
{
	struct pstate_struct* pstate;
	char buf[128];
	
	if (tsk == NULL) {
		return 0;
	}
	
	pstate = alloc_pstate(tsk, current);
	tsk->security = pstate;

	if (tsk->security == NULL) {	
		return error_code;
	}

	sprintf(buf, " creates the process %d\n", tsk->pid);
	print_prs(WAR_INF, current);
	print_log(CTN_INF, buf);

	return 0;
}

void hipp_task_free_security(struct task_struct *tsk)
{
	free_pstate(tsk);
}

void hipp_inode_free_security(struct inode *inode)
{
	struct istate_struct *isec;
	
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [hipp_inode_free_security]\n");
		return;
	}
	
	isec = get_istate(inode);
	if (isec != NULL) {
		isec->magic = -1;
		kfree(isec);
	}
	inode->i_security = NULL;	
}

void hipp_msg_msg_free_security(struct msg_msg* msg)
{
	struct istate_struct *msec;
	
	if (msg == NULL) {
		print_log(ERR_INF, "null error in [hipp_inode_free_security]\n");
		return;
	}
	
	msec = get_mstate(msg);
	if (msec != NULL) {
		msec->magic = -1;
		kfree(msec);
	}
	msg->security = NULL;	
}

void hipp_shm_free_security(struct shmid_kernel *shp)
{
	struct shm_state_struct* shm_state;
	
	if (shp == NULL) {
		print_log(ERR_INF, "null error in [hipp_shm_free_security]\n");
		return;
	}
	
	shm_state = get_shm_state(shp);
	if (shm_state != NULL) {
		shm_state->magic = -1;
		kfree(shm_state);
	}
	shp->shm_perm.security = NULL;
}

int hipp_unix_may_send(struct socket* sock, struct socket* other)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	unix_local_domain_send(pstate, SOCK_INODE(other));	
	return 0;	
}

int hipp_socket_sendmsg(struct socket* sock, struct msghdr* msg, int size)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	struct inode* local_inode;
	struct sock* loopback_sock;

	if (process_sock(sock, DIR_SEND, &local_inode, &loopback_sock) == 1) {
		if (local_inode != NULL) {
			unix_local_domain_send(pstate, local_inode);
		}
		else if (loopback_sock != NULL) {
			//sprintf(buf, "loopback send %u\n", (int)loopback_sock);
			//print_prs(SPE_INF, current);
			//print_log(CTN_INF, buf);
			loopback_send(pstate, loopback_sock);
		}
	}
	return 0;
}

int hipp_socket_recvmsg(struct socket* sock, struct msghdr* msg, int size, int flags)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	struct sock* loopback_sock;
	struct inode* local_inode;

	if (process_sock(sock, DIR_RECV, &local_inode, &loopback_sock) == 1) {
		if (local_inode != NULL) {
			unix_local_domain_recv(pstate, local_inode);
		}
		else if (loopback_sock != NULL) {
			//sprintf(buf, "loopback recv %u\n", (int)loopback_sock);
			//print_prs(SPE_INF, current);
			//print_log(CTN_INF, buf);
			loopback_recv(pstate, loopback_sock);			
		}
		else {
			recv_net(pstate);
		}
	}
	return 0;
}

int hipp_msg_queue_msgsnd(struct msg_queue* msq, struct msg_msg* msg, int msqflg)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();
	
	msg_queue_send(pstate, msg);
	return 0;
}

int hipp_msg_queue_msgrcv(struct msg_queue* msq, struct msg_msg* msg, struct task_struct* target, long type, int mode)
{
	struct pstate_struct* pstate = (struct pstate_struct*)current_pstate();

	msg_queue_recv(pstate, msg);
	return 0;
}

int hipp_shm_shmat(struct shmid_kernel* shp, char __user *shmaddr, int shmflg)
{
	struct pstate_struct* pstate = current_pstate();
	struct shm_state_struct* shm_state;
	struct shm_state_list_struct* t1;
	struct pstate_list_struct* t2;
	
	if (pstate == NULL || shp == NULL) {
		print_log(ERR_INF, "null error in [hipp_shm_shmat]\n");
		return 0;
	}
	
	shm_state = get_shm_state(shp);	
	if (shm_state == NULL) {
		shm_state = alloc_shm_state(shp);
	}
	if (shm_state == NULL) {
		print_log(ERR_INF, "null error in [hipp_shm_shmat]\n");
		return 0;
	}
	
	if (pstate->integrity_state.integrity_level == LOW_INTEGRITY && (shmflg & SHM_RDONLY) == 0) {
		shm_drop(shm_state);
	}
	else if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY && get_lsp_exception(pstate->huid, &pstate->apply_exceptions) == 0) {
		if (shm_state->integrity_level == LOW_INTEGRITY) {
			print_prs(WAR_INF, current);
			print_log(CTN_INF, " drop integrity to low by receiving shared memory.\n");

			pstate_drop(pstate);
		}
		else {
			if ((shmflg & SHM_RDONLY) == 0) {
				t1 = kmalloc(sizeof(struct shm_state_list_struct), GFP_KERNEL);
				if (t1 == NULL) {
					print_log(ERR_INF, "nomem error in [shm_shmat]. \n");
					return 0;
				}
				t1->shm_state = shm_state;	
				list_add_tail(&(t1->list), &(pstate->shm_state_list.list));
			}
			t2 = kmalloc(sizeof(struct pstate_list_struct), GFP_KERNEL);
			if (t2 == NULL) {
				print_log(ERR_INF, "nomem error in [shm_shmat]. \n");
				return 0;
			}
			t2->pstate = pstate;	
			list_add_tail(&(t2->list), &(shm_state->pstate_list.list));
		}
	}
	
	return 0;
}

int hipp_sb_mount(char *dev_name, struct nameidata * nd, char *type, unsigned long flags, void *data)
{
	struct fs_prot_struct* cur;
	char path[256];

	if ((flags & MS_REMOUNT) > 0 || (flags & MS_BIND) > 0) {
		return 0;
	}

	if (type == NULL || dev_name == NULL) {
		print_log(ERR_INF, "null error in [hipp_sb_mount]\n");
		return 0;
	}
	
	if (is_special_fs_type(type) == 1) {
		if (get_absolute_path_by_nd(nd, path) != 0 || check_special_fs_type(type, path) != 1) {

			print_prs(EXP_INF, current);
			sprintf(path, "failed to mount the file system %s\n", dev_name);
			print_log(CTN_INF, path);
			return error_code;
		}
		return 0;
	}
	
	list_for_each_entry(cur, &fs_prot_list->list, list) {
		if (strcmp(cur->dev_name, dev_name) == 0) {
			if (current->namespace != init_namespace) {

				print_prs(EXP_INF, current);
				sprintf(path, "failed to mount the file system %s\n", dev_name);
				print_log(CTN_INF, path);
				return error_code;
			}
			if (get_absolute_path_by_nd(nd, path) != 0 || strcmp(cur->mount_point, path) != 0) {

				print_prs(EXP_INF, current);
				sprintf(path, "failed to mount the file system %s\n", dev_name);
				print_log(CTN_INF, path);
				return error_code;	
			}
		}
	}
	return 0;
}

int hipp_task_kill(struct task_struct* p, struct siginfo* info, int sig)
{
	struct pstate_struct* pstate = current_pstate();
	struct pstate_struct* p2;
	char buf[128];
	
	if (pstate == NULL || p == NULL || info == NULL) {
		print_log(ERR_INF, "null error in [hipp_task_kill]\n");
		return 0;
	}
	p2 = get_pstate(p);
	
	if (p2 == NULL) {
		print_log(ERR_INF, "null error in [hipp_task_kill]\n");
		return 0;
	}
	
	if (info != SEND_SIG_NOINFO && (info <= SEND_SIG_FORCED || SI_FROMKERNEL(info))) {
		return 0;
	}
	
	if (sig >= 32) {
		if (pstate->path != NULL && p2->path != NULL && strcmp(pstate->path, p2->path) == 0) {
			return 0;
		}
		if (pstate->path != NULL) {
			print_log(EXP_INF, pstate->path);
		}
		if (p2->path != NULL) {
			print_log(EXP_INF, p2->path);
		}
	}

	if (pstate->integrity_state.integrity_level == LOW_INTEGRITY && p2->integrity_state.integrity_level == HIGH_INTEGRITY) {
		print_prs(EXP_INF, current);
		sprintf(buf, "try to send signal %d to", sig);
		print_log(CTN_INF, buf);
		print_prs(CTN_INF, p);
		print_log(CTN_INF, "\n");
		
		return error_code;
	}
	return 0;
}

void set_security_operations(void)
{
	struct security_operations *ops = &hipp_ops;

	memset(ops, 0, sizeof(struct security_operations));

	ops->ptrace = cap_ptrace;
	ops->capget = cap_capget;
	ops->capset_check = cap_capset_check;
	ops->capset_set = cap_capset_set;
	ops->settime = cap_settime,
	ops->netlink_send = cap_netlink_send;
	ops->netlink_recv = cap_netlink_recv;
	ops->bprm_set_security = cap_bprm_set_security,
	ops->bprm_secureexec = cap_bprm_secureexec,
	ops->task_reparent_to_init = cap_task_reparent_to_init,
	ops->syslog = cap_syslog,
	ops->vm_enough_memory = cap_vm_enough_memory;

	ops->bprm_apply_creds = hipp_bprm_apply_creds;	
	ops->task_alloc_security = hipp_task_alloc_security;
	ops->task_free_security = hipp_task_free_security;

	ops->task_setuid = hipp_task_setuid;
	ops->task_setgid = hipp_task_setgid;
	ops->task_post_setuid = hipp_task_post_setuid,
	ops->capable = hipp_capable;

	ops->inode_create = hipp_inode_create;
	ops->inode_link = hipp_inode_link;
	ops->inode_unlink = hipp_inode_unlink;
	ops->inode_symlink = hipp_inode_symlink;
	ops->inode_mkdir = hipp_inode_mkdir;
	ops->inode_rmdir = hipp_inode_rmdir;
	ops->inode_rename = hipp_inode_rename;
	ops->inode_permission = hipp_inode_permission;
	ops->inode_setattr = hipp_inode_setattr;
	ops->inode_setxattr = hipp_inode_setxattr;
	ops->inode_removexattr = hipp_inode_removexattr;
	ops->file_permission = hipp_file_permission;
	ops->file_ioctl = hipp_file_ioctl;
	ops->file_mmap = hipp_file_mmap;
	ops->file_fcntl = hipp_file_fcntl;
	ops->file_mprotect = hipp_file_mprotect;

	ops->inode_free_security = hipp_inode_free_security;
	ops->msg_msg_free_security = hipp_msg_msg_free_security;
	ops->shm_free_security = hipp_shm_free_security;
	ops->socket_recvmsg = hipp_socket_recvmsg;
	ops->socket_sendmsg = hipp_socket_sendmsg;
	ops->unix_may_send = hipp_unix_may_send;
	ops->msg_queue_msgsnd = hipp_msg_queue_msgsnd;
	ops->msg_queue_msgrcv = hipp_msg_queue_msgrcv;
	ops->shm_shmat = hipp_shm_shmat;
	ops->sb_mount = hipp_sb_mount;
	//ops->task_kill = hipp_task_kill;
}
