#include "pstate.h"
#include "printlog.h"
#include "apply_exception.h"
#include "securitycheck.h"

int is_valid_pstate(struct pstate_struct* pstate) 
{
	if (pstate == NULL) {
		return 0;
	}
	
	if (pstate->magic != HIPP_MAGIC) {
		return 0;
	}
	
	return 1;
}

int is_valid_istate(struct istate_struct* istate) 
{
	if (istate == NULL) {
		return 0;
	}
	
	if (istate->magic != HIPP_MAGIC) {
		return 0;
	}
	
	return 1;
}

int is_valid_shm_state(struct shm_state_struct* shm_state)
{
	if (shm_state == NULL) {
		return 0;
	}
	
	if (shm_state->magic != HIPP_MAGIC) {
		return 0;
	}
	
	return 1;
}

void release_apply_exceptions(struct apply_exception_struct* es)
{
	struct list_head *pos, *q;
	struct apply_exception_struct* e;
	
	if (es == NULL) {
		print_log(ERR_INF, "null error in [release_apply_exceptions]\n");
		return;
	}
	
	list_for_each_safe(pos, q, &es->list){
		 e = list_entry(pos, struct apply_exception_struct, list);
		 list_del(pos);
		 kfree(e);
	}
}

int copy_apply_exceptions(struct apply_exception_struct* a2, struct apply_exception_struct* a1)
{
	struct apply_exception_struct* t1;
	struct apply_exception_struct* t2;
	
	if (a1 == NULL || a2 == NULL) {
		print_log(ERR_INF, "null error in [copy_apply_exceptions].\n");
		return -1;
	}
	
	INIT_LIST_HEAD(&a2->list);
	list_for_each_entry(t1, &a1->list, list) {
		t2 = kmalloc(sizeof(struct apply_exception_struct), GFP_KERNEL);
		if (t2 == NULL) {
			print_log(ERR_INF, "nomem error in [copy_apply_exceptin]\n");
			return -1;
		}
		t2->exception = t1->exception;
		list_add_tail(&(t2->list), &(a2->list));
	}
	
	return 0;
}

int print_apply_exception(struct apply_exception_struct* aes)
{
	struct apply_exception_struct* ae;
	struct exception_struct* e;
	struct file_exception_struct* fe;
	struct exec_exception_struct* ee;
	char buf[256];
	
	if (aes == NULL) {
		print_log(ERR_INF, "null error in [print_apply_exception].\n");
		return 0;
	}
	
	list_for_each_entry(ae, &aes->list, list) {
		if (ae->exception != NULL) {
			e = ae->exception;
			sprintf(buf, "\t %s %d %d %d %u\n", e->path, e->rap, e->fpp, e->lsp, e->cap);
			print_log(PGS_INF, buf);
			list_for_each_entry(fe, &e->file_exception.list, list) {
				sprintf(buf, "\t file: %s %d %d\n", fe->path, fe->mode, fe->effect);
				print_log(SIM_INF, buf);	
			}
			list_for_each_entry(ee, &e->exec_exception.list, list) {
				sprintf(buf, "\t exec: %s\n", ee->path);
				print_log(SIM_INF, buf);	
			}
			print_log(SIM_INF, "\t #\n");			
		}
	}
	return 0;
}

int establish_exception(struct apply_exception_struct* aes, char* new_bin)
{
	struct exception_struct* e;
	struct apply_exception_struct* ae;
	struct account_list_struct* acc;
	
	if (aes == NULL || new_bin == NULL || global_exception == NULL) {
		print_log(ERR_INF, "null error in [establish_exception].\n");
		return -1;
	}
	
	INIT_LIST_HEAD(&aes->list);
	list_for_each_entry(e, &global_exception->list, list) {
		if (strcmp(new_bin, e->path) == 0 || strcmp(e->path, "*") == 0) {
			ae = kmalloc(sizeof(struct apply_exception_struct), GFP_KERNEL);
			if (ae == NULL) {
				print_log(ERR_INF, "nomem error in [copy_apply_exceptin]\n");
				return -1;
			}
			ae->exception = e;
			list_add_tail(&(ae->list), &(aes->list));
		}
	}
	
	list_for_each_entry(acc, &account_list->list, list) {
		if (acc->user_exception != NULL) {
			list_for_each_entry(e, &acc->user_exception->list, list) {
				if (strcmp(new_bin, e->path) == 0 || strcmp(e->path, "*") == 0) {
					ae = kmalloc(sizeof(struct apply_exception_struct), GFP_KERNEL);
					if (ae == NULL) {
						print_log(ERR_INF, "nomem error in [copy_apply_exceptin]\n");
						return -1;
					}
					ae->exception = e;
					list_add_tail(&(ae->list), &(aes->list));
				}
			}	
		}
	}
	return 0;
}

int initialize_integrity_state(struct integrity_state_struct* i1, struct integrity_state_struct* i2)
{
	if (i1 == NULL || i2 == NULL) {
		print_log(ERR_INF, "null error in [initialize_integrity_state].\n");
		return -1;
	}
	i1->integrity_level = i2->integrity_level;
	i1->mid_user_id = i2->mid_user_id;
	return 0;
}

struct pstate_struct* alloc_pstate(struct task_struct* tsk, struct task_struct* parent_tsk)
{
	struct pstate_struct* pstate;
	struct pstate_struct* parent_pstate = NULL;
	int k;

	if (tsk == NULL) {
		print_log(ERR_INF, "null error in [current_stat]\n");
		return NULL;
	}

	pstate = kmalloc(sizeof(struct pstate_struct), GFP_KERNEL);
	if (pstate == NULL) {
		print_log(ERR_INF, "nomem error in [alloc_pstate].\n");
		return NULL;
	}
	
	pstate->magic = HIPP_MAGIC;
	pstate->tsk = tsk;
	pstate->path = NULL;
	pstate->huid = 0;
	pstate->integrity_state.integrity_level = HIGH_INTEGRITY;
	pstate->integrity_state.mid_user_id = 0;
	INIT_LIST_HEAD(&pstate->shm_state_list.list);
	
	if (parent_tsk == NULL && tsk->parent != tsk) {
		parent_tsk = tsk->parent;
	}
	if (parent_tsk == tsk) {
		parent_tsk = NULL;
	}
	parent_pstate = get_pstate(parent_tsk);
	if (parent_pstate == NULL) {
		pstate->huid = get_huid(tsk);
		INIT_LIST_HEAD(&(pstate->apply_exceptions.list));
		print_log(ERR_INF, "not parent found in [alloc_pstate]\n");
	}
	else {
		pstate->huid = parent_pstate->huid;
		k = get_huid(tsk);
		pstate->huid = update_huid(pstate->huid, k); 
		if (initialize_integrity_state(&pstate->integrity_state, &parent_pstate->integrity_state) != 0) {
			kfree(pstate);
			return NULL;
		}
		if (parent_pstate->path != NULL) {
			pstate->path = kmalloc(strlen(parent_pstate->path) + 1, GFP_KERNEL);
			if (pstate->path == NULL) {
				print_log(ERR_INF, "nomem error in [alloc_pstate]\n");
				kfree(pstate);
				return NULL;
			}
			strcpy(pstate->path, parent_pstate->path);
		}
		//INIT_LIST_HEAD(&(pstate->apply_exceptions.list));
		
		if (copy_apply_exceptions(&pstate->apply_exceptions, &parent_pstate->apply_exceptions) != 0) {
			kfree(pstate->path);
			kfree(pstate);
			return NULL;
		}
		
	}
	return pstate;
}

void free_pstate(struct task_struct* tsk)
{
	struct pstate_struct* pstate;
	
	if (tsk == NULL) {
		print_log(ERR_INF, "null error in [free_pstate]");
		return;
	}
	
	pstate = (struct pstate_struct*)tsk->security;
	if (is_valid_pstate(pstate) == 1) {
		pstate->magic = -1;
		if (pstate->path != NULL) {
			kfree(pstate->path);
		}
		//release_shm_list(&pstate->shm_state_list);
		release_apply_exceptions(&pstate->apply_exceptions);
		kfree(pstate);
	}
	tsk->security = NULL;	
}

struct istate_struct* alloc_istate(struct inode* inode)
{
	struct istate_struct* isec;
	if (inode == NULL) {
		return NULL;
	}
	
	isec = kmalloc(sizeof(struct istate_struct), GFP_KERNEL);
	if (isec == NULL) {
		print_log(ERR_INF, "nomem error in [alloc_istate]\n");
		return NULL;
	}
	isec->magic = HIPP_MAGIC;
	isec->integrity_level = HIGH_INTEGRITY;
	isec->mid_user_id = 0;
	inode->i_security = isec;
	return isec;
}

struct pstate_struct* get_pstate(struct task_struct* tsk) 
{
	struct pstate_struct* pstate;
	
	if (tsk == NULL) {
		return NULL;
	}
	pstate = (struct pstate_struct*)(tsk->security);
	if (is_valid_pstate(pstate) == 0) {
		pstate = (struct pstate_struct*)alloc_pstate(tsk, NULL);
		tsk->security = pstate;
	}
	return (struct pstate_struct*)pstate;
}

struct pstate_struct* current_pstate(void) 
{
	struct pstate_struct* pstate;
	
	if (current == NULL) {
		print_log(ERR_INF, "null error in [current_stat]\n");
		return NULL;
	}
	pstate = get_pstate(current);
	return pstate;
}

int exec_program(struct pstate_struct* pstate, char* new_bin, struct dentry* dentry)
{
	int k = -1;
	
	if (pstate == NULL || new_bin == NULL) {
		return -1;
	}
	
	if (dentry != NULL && dentry->d_inode != NULL) {
		k = get_integrity_level(dentry->d_inode);
	}
	
	if (k != LOW_INTEGRITY && 
	   (pstate->integrity_state.integrity_level == HIGH_INTEGRITY 
	    || get_exec_exception(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, new_bin) == 1)) {
		release_apply_exceptions(&pstate->apply_exceptions);

		if (pstate->path != NULL) {
			kfree(pstate->path);
		}
		pstate->path = kmalloc(strlen(new_bin) + 1, GFP_KERNEL);
		if (pstate->path == NULL) {
			print_log(ERR_INF, "nomem error in [exec_program]. \n");
			return -1;
		}
		strcpy(pstate->path, new_bin);
		if (establish_exception(&pstate->apply_exceptions, new_bin) != 0) {
			return -1;
		}
	
		//print_prs(WAR_INF, current);
		//print_log(CTN_INF, " exception policy applied.\n");
	}
	else if (k != LOW_INTEGRITY && pstate->integrity_state.integrity_level == LOW_INTEGRITY &&
	         get_execinh_exception(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, new_bin) == 1) {
	         
	}
	else {
		release_apply_exceptions(&pstate->apply_exceptions);
	}
	return 0;
}

struct istate_struct* get_istate(struct inode* inode)
{
	struct istate_struct* isec;
	
	if (inode == NULL || inode->i_security == NULL) {
		return NULL;
	}
	isec = (struct istate_struct*)inode->i_security;
	if (is_valid_istate(isec) == 0) {
		return NULL;
	}
	return isec;
}

struct istate_struct* get_mstate(struct msg_msg* msg)
{
	struct istate_struct* msec;
	
	if (msg == NULL || msg->security == NULL) {
		return NULL;
	}
	msec = (struct istate_struct*)msg->security;
	if (is_valid_istate(msec) == 0) {
		return NULL;
	}
	return msec; 
}

struct shm_state_struct* get_shm_state(struct shmid_kernel* shp)
{
	struct shm_state_struct* shm_state;
	
	if (shp == NULL) {
		return NULL;
	}
	shm_state = (struct shm_state_struct*)(shp->shm_perm.security);
	if (is_valid_shm_state(shm_state) == 0) {
		return NULL;
	}
	return shm_state;
}

struct istate_struct* get_sstate(struct sock* sk)
{
	struct istate_struct* sstate;
	
	if (sk == NULL || sk->sk_security == NULL) {
		return NULL;
	}
	sstate = (struct istate_struct*)sk->sk_security;
	if (is_valid_istate(sstate) == 0) {
		return NULL;
	}
	return sstate;
}

struct istate_struct* alloc_sstate(struct sock* sk)
{
	struct istate_struct* sstate;
	
	if (sk == NULL) {
		return NULL;
	}
	
	sstate = kmalloc(sizeof(struct istate_struct), GFP_KERNEL);
	if (sstate == NULL) {
		print_log(ERR_INF, "nomem error in [alloc_sstate]");
		return NULL;
	}
	sstate->magic = HIPP_MAGIC;
	sstate->integrity_level = HIGH_INTEGRITY;
	sstate->mid_user_id = 0;
	sk->sk_security = sstate;
	return sstate;
}

struct istate_struct* alloc_mstate(struct msg_msg* msg)
{
	struct istate_struct* msec;
	
	if (msg == NULL) {
		return NULL;
	}
	
	msec = kmalloc(sizeof(struct istate_struct), GFP_KERNEL);
	if (msec == NULL) {
		print_log(ERR_INF, "nomem error in [alloc_mstate]\n");
		return NULL;
	}
	msec->magic = HIPP_MAGIC;
	msec->integrity_level = HIGH_INTEGRITY;
	msec->mid_user_id = 0;
	msg->security = msec;
	return msec;
}

struct shm_state_struct* alloc_shm_state(struct shmid_kernel* shp)
{
	struct shm_state_struct* shm_state;
	
	if (shp == NULL) {
		return NULL;
	}
	
	shm_state = kmalloc(sizeof(struct shm_state_struct), GFP_KERNEL);
	if (shm_state == NULL) {
		print_log(ERR_INF, "nomem error in [alloc_shm_state]\n");
		return NULL;
	}
	shm_state->magic = HIPP_MAGIC;
	shm_state->integrity_level = HIGH_INTEGRITY;
	shm_state->mid_user_id = 0;
	INIT_LIST_HEAD(&(shm_state->pstate_list.list));
	shp->shm_perm.security = shm_state;
	return shm_state;
}

int get_huid(struct task_struct* tsk)
{
	int a = 0;
	
	if (tsk == NULL) {
		print_log(ERR_INF, "null error in [get_huid]\n");
		return 0;
	}
	
	if (tsk->uid >= 500) {
		if (a == 0) {
			a = tsk->uid;
		}
		else if (tsk->uid != a) {
			a = -1;
		}
	}

	if (tsk->euid >= 500) {
		if (a == 0) {
			a = tsk->euid;
		}
		else if (tsk->euid != a) {
			a = -1;
		}
	}

	if (tsk->suid >= 500) {
		if (a == 0) {
			a = tsk->suid;
		}
		else if (tsk->suid != a) {
			a = -1;
		}
	}
	
	return a;
}

int update_huid(int a, int b)
{
	if (a == -1 || b == -1) {
		return -1;
	}
	
	if (a == 0) {
		return b;
	}
	else if (b == 0) {
		return a;
	}
	else if (a != b) {
		return -1;
	}
	
	return a;
}

int recv_net(struct pstate_struct* pstate)
{
//	return 0;
	if (pstate == NULL) {
		print_log(ERR_INF, "null error in [recv_net]\n");
		return -1;
	}
	
	if (get_rap_exception(pstate->huid, &pstate->apply_exceptions) == 1) {
		return 0;
	}
	
	if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
		print_prs(WAR_INF, current);
		print_log(CTN_INF, " drop integrity to low by receiving network.\n");

		pstate_drop(pstate);	
	}
	return 0;
}
