#include "securitycheck.h"
#include "printlog.h"
#include "apply_exception.h"
#include "utilities.h"
#include "pstate.h"
#include "filesystem.h"

int has_integrity(struct inode* inode)
{
	if (inode == NULL || S_ISDIR(inode->i_mode) || S_ISLNK(inode->i_mode) || S_ISSOCK(inode->i_mode)) {
		return 0;
	}
	if (inode->i_ino == dev_ino || inode->i_ino == tty_ino) {
		return 0;
	}
	if (is_fs_type(inode, "devpts") == 1) {
		return 0;
	}
	return 1;
}

int use_security_field(struct inode* inode)
{
	char type[32];
	
	if (inode == NULL) {
		return 0;
	}
	
	if (S_ISFIFO(inode->i_mode)) {
		return 1;
	}
	
	get_fs_type_by_inode(inode, type);
	if (is_special_fs_type(type) == 1) {
		return 1;
	}

	return 0;
}

int get_read_protected_level(struct inode* inode)
{
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [get_read_protected_level].\n");
		return UN_PROTECTED;
	}
	
	if (is_protected(inode) == 0) {
		return UN_PROTECTED;
	}
	
	if (inode->i_uid >= 500 || is_world_readable(inode) == 1) {
		return LOW_READ_PROTECTED;
	}
	else {
		return HIGH_READ_PROTECTED;
	}
}

int get_write_protected_level(struct inode* inode)
{
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [get_write_protected_level].\n");
		return UN_PROTECTED;
	}
	
	if (is_protected(inode) == 0) {
		return UN_PROTECTED;
	}
	
	if (is_world_writable(inode) == 1) {
		return LOW_WRITE_PROTECTED;
	}
	else {
		return HIGH_WRITE_PROTECTED;
	}
}

int get_prot_level_by_inode(struct inode* inode)
{
	struct vfsmount* vfsmount = NULL;
	int k;
	char type[32];
	//char buf[128];

	if (inode == NULL || inode->i_sb == NULL) {
		return FS_NO_MOUNT_POINT;
	}
	
	get_fs_type_by_inode(inode, type);
	//sprintf(buf, "fs type: %s\n", type);
	//print_log(SIM_INF, buf);
	if (is_special_fs_type(type) == 1) {
		return FS_SPECIAL;
	}
	
	vfsmount = get_vfsmount_by_sb(inode->i_sb, init_namespace);
	
	if (vfsmount != NULL && vfsmount->mnt_devname != NULL) {	
		//sprintf(buf, "dev_name: %s\n", vfsmount->mnt_devname);
		//print_log(SIM_INF, buf);
		
		k = get_fs_prot_level(vfsmount->mnt_devname);
		return k;
	}
	else {
		return FS_NO_MOUNT_POINT;
	}
}

int is_protected(struct inode* inode)
{
	int k;
	
	if (inode == NULL) {
		return 0;
	}
	
	k = get_prot_level_by_inode(inode);

	if (k == FS_NO_MOUNT_POINT || k ==FS_UNPROTECTED) {
		return 0;
	}

	if (is_fs_type(inode, "tmpfs") || is_fs_type(inode, "pipefs") || is_fs_type(inode, "devpts")) {
		return 0;
	}
	
	if (S_ISLNK(inode->i_mode) || S_ISSOCK(inode->i_mode) || S_ISFIFO(inode->i_mode)) {
		return 0;
	}
	
	return 1;
}

int get_integrity_level(struct inode* inode)
{
	int k;
	struct istate_struct* isec;
	
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [get_integrity_level].\n");
		return NO_INTEGRITY;
	}
	
	if (has_integrity(inode) == 0) {
		return NO_INTEGRITY;
	}

	if (use_security_field(inode) == 1) {
		isec = get_istate(inode);
		if (isec == NULL) {
			return HIGH_INTEGRITY;
		}
		else {
			return isec->integrity_level;
		}		
	}
	
	k = get_prot_level_by_inode(inode);
	if (k == FS_PROTECTED || k == FS_UNPROTECTED || k == FS_NO_MOUNT_POINT) {
		return LOW_INTEGRITY;
	}
	else if (k == FS_TRUSTED) {
		if (is_world_writable(inode) == 1 || is_sticky_set(inode) == 1) {
			return LOW_INTEGRITY;
		}
		else {
			return HIGH_INTEGRITY;
		}
	}

	return NO_INTEGRITY;
}

int read_check(struct dentry* dentry, struct pstate_struct* pstate)
{
	struct inode* inode;
	int k;
	int r = 0;
	int m;
	
	if (dentry == NULL || pstate == NULL || dentry->d_inode == NULL) {
		print_log(ERR_INF, "null error in [read_check]\n");
		return 0;
	}
	
	inode = dentry->d_inode;	
	k = get_read_protected_level(inode);
	if (k == HIGH_READ_PROTECTED) {
		if (pstate->integrity_state.integrity_level != HIGH_INTEGRITY) {
			r = (r | DO_READ);
		}
	}
	
	if (r != 0) {
		m = get_read_exception_by_dentry(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, dentry);
		if (m == 0) {
			return -1;
		}
	}
	return 0;
}

int write_check(struct dentry* dentry, struct pstate_struct* pstate)
{
	struct inode* inode;
	int k;
	int r = 0;
	int m;
	
	if (dentry == NULL || pstate == NULL || dentry->d_inode == NULL) {
		print_log(ERR_INF, "null error in [write_check]\n");
		return 0;
	}
	
	inode = dentry->d_inode;	
	k = get_write_protected_level(inode);
	if (k == HIGH_WRITE_PROTECTED) {
		if (pstate->integrity_state.integrity_level != HIGH_INTEGRITY) {
			r = (r | DO_WRITE);
		}
	}
	
	if (r != 0) {
		m = get_write_exception_by_dentry(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, dentry);
		if (m == 0) {
			return -1;
		}
	}
	return 0;
}

int create_check(struct dentry* dentry, struct pstate_struct* pstate)
{
	int m;
	
	if (dentry == NULL || pstate == NULL) {
		print_log(ERR_INF, "null error in [create_check]\n");
		return 0;
	}
	
	if (write_check(dentry->d_parent, pstate) == 0) {
		return 0;
	}
	
	m = get_write_exception_by_dentry(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, dentry);
	if (m == 1) {
		return 0;
	}
	
	return -1;
}

int unlink_check(struct dentry* dentry, struct pstate_struct* pstate)
{
	if (dentry == NULL || pstate == NULL || dentry->d_inode == NULL) {
		print_log(ERR_INF, "null error in [unlink_check]\n");
		return 0;
	}
	
	if (write_check(dentry, pstate) == 0 || write_check(dentry->d_parent, pstate) == 0) {
		return 0;
	}
	
	return -1;
}

int create_link_check(struct dentry* dentry_from, struct dentry* dentry_to, struct pstate_struct* pstate)
{
	if (dentry_from == NULL || pstate == NULL || dentry_from->d_inode == NULL || dentry_to == NULL) {
		print_log(ERR_INF, "null error in [create_link_check]\n");
		return 0;
	}
	
	if (read_check(dentry_from, pstate) == 0 && write_check(dentry_from, pstate) == 0 && create_check(dentry_to, pstate) == 0) {
		return 0;
	}
	
	return -1;
}

int rename_check(struct dentry* dentry_from, struct dentry* dentry_to, struct pstate_struct* pstate)
{
	if (dentry_from == NULL || pstate == NULL || dentry_from->d_inode == NULL || dentry_to == NULL) {
		print_log(ERR_INF, "null error in [rename_check]\n");
		return 0;
	}
	
	if (create_link_check(dentry_from, dentry_to, pstate) == 0 && unlink_check(dentry_from, pstate) == 0) {
		return 0;
	}
	
	return -1;
}

int exec_check(struct dentry* dentry, struct pstate_struct* pstate)
{
	if (dentry == NULL || pstate == NULL || dentry->d_inode == NULL) {
		print_log(ERR_INF, "null error in [exec_check]\n");
		return 0;
	}
	
	if (S_ISDIR(dentry->d_inode->i_mode) != 0) {
		return 0;
	}
	
	if (read_check(dentry, pstate) != 0) {
		return -1;
	}
	return 0;
}

int capability_check(int cap, struct pstate_struct* pstate)
{
	if (cap < 0 || cap >= 32 || pstate == NULL) {
		print_log(ERR_INF, "null error in [capability_chec]\n");
		return 0;
	}
	
	if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
		return 0;
	}
	
	if (cap == 6 || cap == 7) {
		return 0;
	}
	
	if (get_cap_exception(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, cap) == 1) {
		return 0;
	}

	return -1;
}

int set_permissions_check(struct dentry* dentry, struct pstate_struct* pstate, int new_mode, int new_uid, int* sticky)
{
	struct inode* inode;
	int a, b;
	int r = 0;
	int m;
	
	if (dentry == NULL || pstate == NULL || dentry->d_inode == NULL || sticky == NULL) {
		return -1;
	}
	*sticky = 0;
	
	inode = dentry->d_inode;
	
	a = get_read_protected_level(inode);
	if (new_uid >= 500 || (new_mode & S_IROTH) > 0) {
		b = LOW_READ_PROTECTED;
	}
	else {
		b = HIGH_READ_PROTECTED;
	}
	
	if (a == HIGH_READ_PROTECTED && b == LOW_READ_PROTECTED) {
		r = (r | DO_READ);
	}
	
	a = get_write_protected_level(inode);
	if ((new_mode & S_IWOTH) > 0) {
		b = LOW_WRITE_PROTECTED;
	}
	else {
		b = HIGH_WRITE_PROTECTED;
	}

	if (a == LOW_WRITE_PROTECTED && b == HIGH_WRITE_PROTECTED) {
		r = (r | DO_WRITE);
		*sticky = 1;
	}

	if (a == HIGH_WRITE_PROTECTED) {
		r = (r | DO_WRITE);
	}
	
	a = is_sticky_set(inode);
	if ((new_mode & S_ISVTX) > 0) {
		b = 1;
	}
	else {
		b = 0;
	}
	
	if (a == 1 && b == 0) {
		if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY && strcmp(pstate->path, UPGRADE_INTEGRITY_UTILITY_NAME) == 0) {
			
		}
		else {
			*sticky = 1;
		}
	}
	
	if (a == 0 && b == 1) {
		r = (r | DO_WRITE);
	}
	
	if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
		r = 0;
	}
	
	if ((r & DO_READ) > 0) {
		m = get_read_exception_by_dentry(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, dentry);
		if (m == 0) {
			return -1;
		}		
	}
	
	if ((r & DO_WRITE) > 0) {
		m = get_write_exception_by_dentry(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, dentry);
		if (m == 0) {
			return -1;
		}		
	}
	return 0;
}

int read_from(struct dentry* dentry, struct pstate_struct* pstate)
{
	struct inode* inode;
	int k;
	
	if (dentry == NULL || dentry->d_inode == NULL || pstate == NULL) {
		print_log(ERR_INF, "null error in [read_from].\n");
		return -1;
	}
	
	inode = dentry->d_inode;
	k = get_integrity_level(inode);
	if (k == LOW_INTEGRITY) {
		if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
			if (get_fpp_exception(pstate->huid, pstate->path, &pstate->apply_exceptions) == 0) {
				if (get_lsp_exception(pstate->huid, &pstate->apply_exceptions) == 0) {
					print_prs(WAR_INF, current);
					print_log(CTN_INF, " drop integrity to low by reading ");
					print_den(CTN_INF, dentry);
					print_log(CTN_INF, "\n");
	
					pstate_drop(pstate);
				}
			}
		}
	}
	return 0;
}

int exec_from(struct dentry* dentry, struct pstate_struct* pstate)
{
	struct inode* inode;
	int k;
	if (dentry == NULL || dentry->d_inode == NULL || pstate == NULL) {
		print_log(ERR_INF, "null error in [exec_from].\n");
		return -1;
	}
	
	inode = dentry->d_inode;
	k = get_integrity_level(inode);
	if (k == LOW_INTEGRITY) {
		if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
			print_prs(WAR_INF, current);
			print_log(CTN_INF, " drop integrity to low by execing ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");

			pstate_drop(pstate);
		}
	}
	return 0;
}

int write_to(struct dentry* dentry, struct pstate_struct* pstate)
{
	struct inode* inode;
	struct istate_struct* isec;
	int k;
	if (dentry == NULL || dentry->d_inode == NULL || pstate == NULL) {
		print_log(ERR_INF, "null error in [write_to].\n");
		return -1;
	}
	
	inode = dentry->d_inode;
	k = get_integrity_level(inode);
	if (k == HIGH_INTEGRITY && pstate->integrity_state.integrity_level == LOW_INTEGRITY) {
		if (get_drop_exception(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, dentry) == 0) {
			if (use_security_field(inode) == 1) {
				isec = get_istate(inode);
				if (isec == NULL) {
					isec = alloc_istate(inode);					
				}
				if (isec == NULL) {
					return 0;
				}
				isec->integrity_level = LOW_INTEGRITY;
			}
			else {
				set_sticky_bit(inode);
			}		
			print_prs(WAR_INF, current);
			print_log(CTN_INF, " drop the integrity of inode ");
			print_den(CTN_INF, dentry);
			print_log(CTN_INF, "\n");
		}
	}
	
	return 0;
}

void unix_local_domain_send(struct pstate_struct* pstate, struct inode* inode)
{
	struct istate_struct* isec;
	char buf[64];

	if (pstate == NULL || inode == NULL) {
		print_log(ERR_INF, "null error in [unix_local_domain_send]\n");
		return;
	}
	
	if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
		return;
	}
	
	isec = get_istate(inode);
	if (isec == NULL) {
		isec = alloc_istate(inode);					
	}
	if (isec == NULL) {
		return;
	}

	if (isec->integrity_level == HIGH_INTEGRITY) {
		isec->integrity_level = LOW_INTEGRITY;
		
		print_prs(WAR_INF, current);
		sprintf(buf, " drop the integrity of unix_domain socket %lu\n", inode->i_ino);
		print_log(CTN_INF, buf);
	}
}

void unix_local_domain_recv(struct pstate_struct* pstate, struct inode* inode)
{
	struct istate_struct* isec;
	char buf[64];

	if (pstate == NULL || inode == NULL) {
		print_log(ERR_INF, "null error in [unix_local_domain_send]\n");
		return;
	}
	
	isec = get_istate(inode);
	if (isec != NULL && isec->integrity_level == LOW_INTEGRITY) {
		if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {		
			if (get_lsp_exception(pstate->huid, &pstate->apply_exceptions) == 0) {
				print_prs(WAR_INF, current);
				sprintf(buf, " drop integrity to low by receiving unix_domain socket %lu\n", inode->i_ino);
				print_log(CTN_INF, buf);

				pstate_drop(pstate);
			}
		}
	}
}

void loopback_send(struct pstate_struct* pstate, struct sock* sk)
{
	struct istate_struct* sstate;

	if (pstate == NULL || sk == NULL) {
		print_log(ERR_INF, "null error in [loopback_send]\n");
		return;
	}
	
	if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
		return;
	}
	
	sstate = get_sstate(sk);
	if (sstate == NULL) {
		sstate = alloc_sstate(sk);					
	}
	if (sstate == NULL) {
		return;
	}
	sstate->integrity_level = LOW_INTEGRITY;
}

void loopback_recv(struct pstate_struct* pstate, struct sock* sk)
{
	struct istate_struct* sstate;

	if (pstate == NULL || sk == NULL) {
		print_log(ERR_INF, "null error in [loopback_recv]\n");
		return;
	}
	
	sstate = get_sstate(sk);
	if (sstate != NULL && sstate->integrity_level == LOW_INTEGRITY) {
		if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {		
			if (get_lsp_exception(pstate->huid, &pstate->apply_exceptions) == 0) {	
				print_prs(WAR_INF, current);
				print_log(CTN_INF, " drop integrity to low by receiving loopback.\n");

				pstate_drop(pstate);
			}
		}
	}
}

void msg_queue_send(struct pstate_struct* pstate, struct msg_msg* msg)
{
	struct istate_struct* msec;
	
	if (pstate == NULL || msg == NULL) {
		print_log(ERR_INF, "null error in [msg_queue_send]\n");
		return;
	}
	
	if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
		return;
	}
	
	msec = get_mstate(msg);
	if (is_valid_istate(msec) != 1) {
		msec = alloc_mstate(msg);
		if (msec == NULL) {
			return;
		}
		
		msec->integrity_level = LOW_INTEGRITY;
	}
	else {
		msec->integrity_level = LOW_INTEGRITY;
	}
}

void msg_queue_recv(struct pstate_struct* pstate, struct msg_msg* msg)
{
	struct istate_struct* msec;

	if (pstate == NULL || msg == NULL) {
		print_log(ERR_INF, "null error in [msg_queue_recv]\n");
		return;
	}
	
	msec = get_mstate(msg);
	if (is_valid_istate(msec) == 1 && msec->integrity_level == LOW_INTEGRITY) {
		if (pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {		
			if (get_lsp_exception(pstate->huid, &pstate->apply_exceptions) == 0) {		
				print_prs(WAR_INF, current);
				print_log(CTN_INF, " drop integrity to low by receiving POSIX message.\n");

				pstate_drop(pstate);			
			}
		}
	}
}

int same_uid(int a, int b)
{
	if (a == b)
	{
		return 1;
	}
	if (a < 500 && b < 500)
	{
		return 1;
	}
	return 0;
}

int setuid_check(int id0, int id1, int id2, struct pstate_struct* pstate, int cap)
{
	int p = 0;
	if (pstate == NULL) {
		print_log(ERR_INF, "null error in [setuid_check].\n");
		return -1;
	}

	if (id0 != -1 && !same_uid(id0, current->uid) && !same_uid(id0, current->euid) && !same_uid(id0, current->suid)) {
		p = 1;
	}
	if (id1 != -1 && !same_uid(id1, current->uid) && !same_uid(id1, current->euid) && !same_uid(id1, current->suid)) {
		p = 1;
	}
	if (id2 != -1 && !same_uid(id2, current->uid) && !same_uid(id2, current->euid) && !same_uid(id2, current->suid)) {
		p = 1;
	}
	
	if (p == 1) {
		if (pstate->integrity_state.integrity_level == LOW_INTEGRITY) {
			if (get_cap_exception(pstate->huid, &pstate->integrity_state, &pstate->apply_exceptions, cap) == 1) {
				return 0;
			}
			else {
				return -1;
			}
		}
	}
	
	return 0;
}

int shm_drop(struct shm_state_struct* shm_state)
{
	struct pstate_list_struct* p;
	
	if (is_valid_shm_state(shm_state) != 1) {
		print_log(ERR_INF, "null error in [shm_drop].\n");
		return 0;
	}
	
	if (shm_state->integrity_level == LOW_INTEGRITY) {
		return 0;
	}
	shm_state->integrity_level = LOW_INTEGRITY;
	
	list_for_each_entry(p, &shm_state->pstate_list.list, list) {
		if (is_valid_pstate(p->pstate) == 1) {
			if (p->pstate->integrity_state.integrity_level == HIGH_INTEGRITY) {
				if (get_lsp_exception(p->pstate->huid, &p->pstate->apply_exceptions) == 0) {
					print_prs(WAR_INF, current);
					print_log(CTN_INF, " drop integrity to low by receiving shared memory.\n");

					pstate_drop(p->pstate);
				}
			}
		}
	}
	return 0;
}

int pstate_drop(struct pstate_struct* pstate)
{
	struct shm_state_list_struct* s;
	
	if (is_valid_pstate(pstate) != 1) {
		print_log(ERR_INF, "null error in [pstate_drop].\n");
		return 0;
	}
	
	if (pstate->integrity_state.integrity_level == LOW_INTEGRITY) {
		return 0;
	}
	pstate->integrity_state.integrity_level = LOW_INTEGRITY;
	
	list_for_each_entry(s, &pstate->shm_state_list.list, list) {
		if (is_valid_shm_state(s->shm_state) == 1) {
			shm_drop(s->shm_state);
		}
	}
	return 0;
}
