#include "utilities.h"
#include "printlog.h"
#include "filesystem.h"

void get_fs_type_by_inode(struct inode* inode, char* type)
{
	type[0] = '\0';
	if (inode != NULL && inode->i_sb != NULL && inode->i_sb->s_type != NULL && inode->i_sb->s_type->name != NULL) {
		strcpy(type, inode->i_sb->s_type->name);
	}		
}

int is_fs_type(struct inode* inode, const char* name)
{
	if (inode != NULL && inode->i_sb != NULL && inode->i_sb->s_type != NULL && inode->i_sb->s_type->name != NULL) {
		if (strcmp(inode->i_sb->s_type->name, name) == 0) {
			return 1;
		}
	}
	return 0;
}

struct vfsmount* get_vfsmount_by_sb(struct super_block* sb, struct namespace* ns)
{
	struct vfsmount* mnt = NULL;
	struct vfsmount* result = NULL;

	if (ns == NULL) {
		print_log(ERR_INF, "null error in [get_vfsmount_by_sb]\n");
		return NULL;
	}
	
	list_for_each_entry(mnt, &(ns->root->mnt_list), mnt_list) {
		if (mnt->mnt_sb == sb) {
			result = mnt;
			break;
		}
	}
	return result;
}

int get_single_path_from_dentry(struct dentry* dentry, char* path)
{
	int p = MAX_PATH_LEN - 1;
	
	if (dentry == NULL || path == NULL) {
		print_log(ERR_INF, "null error 1 in [get_single_path_from_dentry]\n");
		return -1;
	}
	
	path[p] = '\0';	
	while (dentry != NULL) {
		int pp;
		
		if (dentry == dentry->d_parent) {
			break;
		}
		if (dentry->d_name.name == NULL) {
			print_log(ERR_INF, "null error 2 in [get_single_path_from_dentry]\n");
			return -1;
		}
		
		pp = p - dentry->d_name.len;
		if (pp <= 0) {
			print_log(ERR_INF, "too long in [get_single_path_from_dentry]\n");
			return -1;
		}
		memcpy(&(path[pp]), dentry->d_name.name, dentry->d_name.len);
		p = pp - 1;
		path[p] = '/';

		dentry = dentry->d_parent;
	}
	memmove(path, &(path[p]), MAX_PATH_LEN - p);
	return 0;		
}

int get_path_from_dentry_uni(struct dentry* dentry, char* path, struct namespace* ns)
{
	struct vfsmount* cur_mnt = NULL;	
	struct vfsmount* next_mnt = NULL;
	char single_path[256];
	int p = MAX_PATH_LEN - 1;
	int pp;
	
	if (dentry == NULL || path == NULL || ns == NULL) {
		print_log(ERR_INF, "null error in [get_path_from_dentry_uni]\n");
		return -1;
	}

	cur_mnt = get_vfsmount_by_sb(dentry->d_sb, ns);
	if (cur_mnt == NULL) {
		return get_single_path_from_dentry(dentry, path);
	}
		
	path[p] = '\0';
	while (1) {
		get_single_path_from_dentry(dentry, single_path);
		pp = p - strlen(single_path);
		if (pp <= 0) {
			print_log(ERR_INF, "too long in [get_path_from_dentry_uni]\n");
			return -1;
		}
		memcpy(&(path[pp]), single_path, strlen(single_path));
		p = pp;
				
		if (cur_mnt->mnt_parent == NULL || cur_mnt->mnt_parent == cur_mnt) {
			break;
		}
		dentry = cur_mnt->mnt_mountpoint;
		next_mnt = cur_mnt->mnt_parent;
		cur_mnt = next_mnt;
	}
	memmove(path, &(path[p]), MAX_PATH_LEN - p);
	return 0;
}

int get_path_from_dentry(struct dentry* dentry, char* path)
{
	return get_path_from_dentry_uni(dentry, path, current->namespace);
}

int get_absolute_path_by_nd(struct nameidata* nd, char* path)
{
	if (nd == NULL) {
		print_log(ERR_INF, "null error in [get_absolute_path_by_nd]\n");
		return -1;
	}
	return get_path_from_dentry_uni(nd->dentry, path, init_namespace);
}

int set_sticky_bit(struct inode* inode)
{
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [set_sticky_bit]\n");
		return -1;
	}
	
	inode->i_mode = (inode->i_mode | S_ISVTX);
	return 0;
}

int is_sticky_set(struct inode* inode)
{
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [is_sticky_set]\n");
		return -1;
	}
	
	if ((inode->i_mode & S_ISVTX) > 0) {
		return 1;
	}
	return 0;
}

int is_world_readable(struct inode* inode)
{
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [is_world_readable]\n");
		return -1;
	}
	
	if ((inode->i_mode & S_IROTH) > 0) {
		return 1;
	}
	return 0;
}

int is_world_writable(struct inode* inode)
{
	if (inode == NULL) {
		print_log(ERR_INF, "null error in [is_world_writable]\n");
		return -1;
	}
	
	if ((inode->i_mode & S_IWOTH) > 0) {
		return 1;
	}
	return 0;
}

struct dentry* get_dentry_from_path(const char* path)
{
	struct nameidata nd;
	int error;
	struct dentry* res = NULL;

	if (path == NULL) {
		return NULL;
	}
	
	error = path_lookup(path, LOOKUP_FOLLOW, &nd);
	if (!error) {
		res = nd.dentry;
		path_release(&nd);
	}
	else {
		print_log(ERR_INF, "transfer path to dentry failed");
		print_log(ERR_INF, path);
		print_log(ERR_INF, "\n");
		
	}
	return res;
}

struct sock* udp_v4_lookup(struct inet_sock* inet_snd)
{
	struct sock *sk, *result = NULL;
	struct hlist_node *node;
	unsigned short hnum = ntohs(inet_snd->dport);
	int badness = -1;

	read_lock(&udp_hash_lock);
	sk_for_each(sk, node, &udp_hash[hnum & (UDP_HTABLE_SIZE - 1)]) {
		struct inet_sock *inet = inet_sk(sk);
		if (inet->num == hnum && !ipv6_only_sock(sk)) {
			int score = (sk->sk_family == PF_INET ? 1 : 0);
			if (inet->rcv_saddr) {
				if (inet->rcv_saddr != inet_snd->daddr)
					continue;
				score+=2;
			}
			if (inet->daddr) {
				if (inet->daddr != inet_snd->saddr)
					continue;
				score+=2;
			}
			if (inet->dport) {
				if (inet->dport != inet_snd->sport)
					continue;
				score+=2;
			}
			if (sk->sk_bound_dev_if) {
				if (sk->sk_bound_dev_if != loopback_dev.ifindex)
					continue;
				score+=2;
			}
			if(score == 9) {
				result = sk;
				break;
			} else if(score > badness) {
				result = sk;
				badness = score;
			}
		}
	}
	read_unlock(&udp_hash_lock);
	return result;
}

struct sock* process_loopback(struct inet_sock* inet, int version)
{
	struct sock* sk = NULL;
	
	if (inet == NULL) {
		print_log(ERR_INF, "null error in [process_loopback]\n");
		return NULL;
	}
	
	if (version == 4) {
		//sk = tcp_v4_lookup(inet->saddr, inet->sport, inet->daddr, inet->dport, loopback_dev.ifindex);
		sk = inet_lookup(&tcp_hashinfo, inet->saddr, inet->sport, inet->daddr, inet->dport, loopback_dev.ifindex);
	}
	else if (version == 6) {
		//sk = tcp_v6_lookup(&(inet->pinet6->saddr), inet->sport, &(inet->pinet6->daddr), inet->dport, loopback_dev.ifindex);		
	}
	else if (version == 44) {
		sk = udp_v4_lookup(inet);
	}
	return sk;
}

int format_addr(unsigned int k, char* p)
{
	unsigned int a = ((k & (((unsigned int)255) << 24)) >> 24);
	unsigned int b = ((k & (((unsigned int)255) << 16)) >> 16);
	unsigned int c = ((k & (((unsigned int)255) << 8)) >> 8);
	unsigned int d = ((k & (((unsigned int)255) << 0)) >> 0);

	sprintf(p, "%u.%u.%u.%u", a, b, c, d);
	return 0;
}

int process_sock(struct socket* sock, int dir, struct inode** local_inode, struct sock** loopback_sock)
{
	int p;
	struct sock* sk;
	struct inet_sock *inet;
	struct inet_sock *other_inet;
	struct unix_sock *usock;
	unsigned int q0, q1, p0, p1, p2, p3, p4, p5, p6, p7;
	char a0[30], a1[30];
	unsigned int r[4];
	char buf[128];
	
	if (sock == NULL || sock->sk == NULL) {
		return 0;
	}
	
	sk = sock->sk;
	p = sk->__sk_common.skc_family;
	
	*local_inode = NULL;
	*loopback_sock = NULL;
	
	if (p == PF_INET) {
		inet = inet_sk(sk);
		if (inet == NULL) {
			return 0;
		}

		p0 = ntohl(inet->saddr);
		p1 = ntohl(inet->daddr);
		q0 = ntohs(inet->sport);
		q1 = ntohs(inet->dport);
		/*
		print_prs(SPE_INF, current);
		format_addr(p0, a0);
		format_addr(p0, a1);
		sprintf(buf, "inet4: DIR: %d %s : %u, %s : %u.\n", dir, a0, q0, a1, q1);
		print_log(CTN_INF, buf);
		*/
		if (p0 == LOCAL_IP && p1 == LOCAL_IP) {
			if (dir == DIR_SEND) {
				if (sk->sk_type == SOCK_STREAM) {			
					*loopback_sock = process_loopback(inet, 4);
				}
				else {
					*loopback_sock = process_loopback(inet, 44);
				}
				if (*loopback_sock == NULL) {
					return 0;
				}
			}
			else if (dir == DIR_RECV && sock->sk != NULL) {
				*loopback_sock = sock->sk;
				if (*loopback_sock == NULL) {
					return 1;
				}
			}
		}
	}
	else if (p == PF_INET6)	{
		inet = inet_sk(sk);
		if (inet == NULL) {
			return 0;
		}
		
		memcpy(r, inet->pinet6->saddr.in6_u.u6_addr32, sizeof(r));
		p3 = ntohl(r[0]); p2 = ntohl(r[1]); p1 = ntohl(r[2]); p0 = ntohl(r[3]);
		memcpy(r, inet->pinet6->daddr.in6_u.u6_addr32, sizeof(r));
		p7 = ntohl(r[0]); p6 = ntohl(r[1]); p5 = ntohl(r[2]); p4 = ntohl(r[3]);
		q0 = ntohs(inet->sport);
		q1 = ntohs(inet->dport);
		/*
		print_prs(SPE_INF, current);
		format_addr(p0, a0);
		format_addr(p0, a1);
		sprintf(buf, "inet6: DIR: %d %s : %u, %s : %u.\n", dir, a0, q0, a1, q1);
		print_log(CTN_INF, buf);
		*/
		if (p0 == LOCAL_IP && p1 == LOCAL_IP) {
			if (dir == DIR_SEND) {			
				if (sk->sk_type == SOCK_STREAM) {			
					*loopback_sock = process_loopback(inet, 6);
				}
				else {
					*loopback_sock = process_loopback(inet, 66);
				}
				if (*loopback_sock == NULL) {
					return 0;
				}
			}
			else if (dir == DIR_RECV && sock->sk != NULL) {
				*loopback_sock = sock->sk;
				if (*loopback_sock == NULL) {
					return 1;
				}
			}
		}
	}
	else if (p == PF_LOCAL) {
		if (dir == DIR_RECV) {			
			*local_inode = SOCK_INODE(sock);
			if (*local_inode == NULL) {
				return 0;
			}
		}
		else if (dir == DIR_SEND) {
			usock = unix_sk(sk);		
			if (usock == NULL || usock->peer == NULL || usock->peer->sk_socket == NULL) {
				return 0;
			}
			*local_inode = SOCK_INODE(usock->peer->sk_socket);
			if (*local_inode == NULL) {
				return 0;
			}
		}
	}
	else if (p == 16) {
		return 0;
	}
	return 1;
}
